/*          
 * Copyright (c) 2013 Carlo Augusto Grazia, Paolo Valente
 * Universita` di Modena e Reggio Emilia
 * All rights reserved 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifdef _KERNEL
#include <math.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/module.h>
#include <net/if.h> /* IFNAMSIZ */
#include <netinet/in.h>
#include <netinet/ip_var.h>     /* ipfw_rule_ref */
#include <netinet/ip_fw.h>  /* flow_id */
#include <netinet/ip_dummynet.h>
#include <netinet/ipfw/dn_heap.h>
#include <netinet/ipfw/ip_dn_private.h>
#include <netinet/ipfw/dn_sched.h>
#else
#include <dn_test.h>
#endif

#define DN_DOUBLE   5 // XXX Where?


struct mbuf *tosend;    /* packet to send -- also flag to enqueue */
struct mbuf *freelist;

/* Qos and MAC-SAL algorithm */
struct dn_alg *p_qos;
struct dn_alg *p_mac;

struct cfg_s {
    int ac;
    char * const *av;

    int Q; /* MAC-SAL buffer size [pkts] */

    int N; /* number of qos-flows */
    int M; /* number of mac-flows */

    const char *name;
    int loops;

    /* running counters */
    uint32_t    _enqueue;
    uint32_t    drop;
    uint32_t    pending;
    uint32_t    dequeue;

    /* generator parameters */
    int lmin, lmax; /* packet len */
    int flows;  /* number of flows */
    int flowsets;   /* number of flowsets */
    int wsum;   /* sum of weights of all flows */
    int max_flow_w; /* max flow weight */
    int max_y;  /* max random number in the generation */
    int cur_y, cur_fs;  /* used in generation, between 0 and max_y - 1 */
    const char *fs_config; /* flowset config */
    int can_dequeue;

    int weight_distrib; /* 
                         * to set automaticaly weights of MAC flows 
                         * as a function of packet losses probability
                         */

    int intgr_th; /*
                   * emulate an integrated scheduler policy 
                   * using a threshold to set flow in Good or Bad state
                   */

    int print_stats_level;
    struct mbuf *tosend;    /* packet to send -- also flag to enqueue */

    struct mbuf *freelist;

    struct mbuf *head, *tail;   /* a simple tailq */

    /* scheduler hooks */
    int (*enq)(struct dn_sch_inst *, struct dn_queue *,
        struct mbuf *);
    struct mbuf * (*deq)(struct dn_sch_inst *);
    /* size of the three fields including sched-specific areas */
    int schk_len;
    int q_len; /* size of a queue including sched fields */
    int si_len; /* size of a sch_inst including sched fields */
    char *q;    /* array of flow queues */
        /* use a char* because size is variable */
    struct dn_fsk *fs;  /* array of flowsets */
    struct dn_sch_inst *si;
    struct dn_schk *sched;

    int single_queue; /*
               * equal to 1 if the scheduler uses just one
               * queue to store the packets of all the flows
               */
} *macsal, *qossched;

/* global Time Worst-case Fair Index (T-WFI) */
double t_wfi;
/* 
 * global Time Worst-case Fair Index (T-WFI) 
 * (good qos flows mapped in bad MAC-SAL flows) 
 */
double t_wfi_interm;

/* Bit Worst-case Fair Index (B-WFI) (MAC-SAL) */
double b_wfi;
/* global Bit Worstcase Fair Index (B-WFI) (QoS) */
double b_wfi_qos;
/* global Bit Worstcase Fair Index (B-WFI) (QoS) */
double b_wfi_qos_interm;

/* global max queueing delay (both QoS and MAC) */
double q_del;
/* macsal max queueing delay (only MAC) */
double q_del_mac_best_flow;   /* lowest p_loss flows */
double q_del_mac_interm_flow; /* interm p_loss flows */

/* Link speed in Kb/sec */
int bw = 54000;

/* Ploss array (throughput) */
float p_loss[] = { 0, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.99, 0.999, 0.9999 };

/* global refresh for regenerate weight after ploss changes */
int recompute_ploss_interval;

/*
 * FI2Q and Q2FI converts from flow_id to dn_queue and back.
 * We cannot easily use pointer arithmetic because it is variable size.
 */
#define FI2Q(c, i)  ((struct dn_queue *)((c)->q + (c)->q_len * (i)))
#define Q2FI(c, q)  (((char *)(q) - (c)->q)/(c)->q_len)

/* drop is a metter of pointers, we do not mantain the controller logic */
int
double_drop(struct cfg_s *c, struct mbuf *m)
{
    m->m_nextpkt = freelist;
    freelist = m;
    return 0;
}

/* interpret a number in human form */
/* copied from main, should be define as a common function shared */
static long
getnum(const char *s, char **next, const char *key)
{
    char *end = NULL;
    long l;

    if (next)   /* default */
        *next = NULL;
    if (s && *s) {
        DX(4, "token is <%s> %s", s, key ? key : "-");
        l = strtol(s, &end, 0);
    } else {
        DX(4, "empty string");
        l = -1;
    }
    if (l < 0) {
        DX(3, "invalid %s for %s", s ? s : "NULL", (key ? key : "") );
        return 0;   // invalid 
    }
    if (!end || !*end)
        return l;
    if (*end == 'n')
        l = -l; /* multiply by n */
    else if (*end == 'K')
        l = l*1000;
    else if (*end == 'M')
        l = l*1000000;
    else if (*end == 'k')
        l = l*1024;
    else if (*end == 'm')
        l = l*1024*1024;
    else if (*end == 'w')
        ;
    else {/* not recognized */
        D("suffix %s for %s, next %p", end, key, next);
        end--;
    }
    end++;
    DX(4, "suffix now %s for %s, next %p", end, key, next);
    if (next && *end) {
        DX(4, "setting next to %s for %s", end, key);
        *next = end;
    }
    return l;
}

/*
 * flowsets are a comma-separated list of
 * weight:maxlen:flows
 * indicating how many flows are hooked to that fs.
 * Both weight and range can be min-max-steps.
 * In a first pass we just count the number of flowsets and flows,
 * in a second pass we complete the setup.
 */
static void
parse_flowsets(struct cfg_s *c, const char *fs, int pass)
{
    char *s, *cur, *next;
    int n_flows = 0, n_fs = 0, wsum = 0;
    int i, j;
    struct dn_fs *prev = NULL;

    DX(4, "--- pass %d flows %d flowsets %d", pass, c->flows, c->flowsets);
    if (pass == 0)
        c->fs_config = fs;
    s = c->fs_config ? strdup(c->fs_config) : NULL;
    if (s == NULL) {
        if (pass == 0)
            D("no fsconfig");
        return;
    }
    for (next = s; (cur = strsep(&next, ","));) {
        char *p = NULL;
        int w, w_h, w_steps, wi;
        int len, len_h, l_steps, li;
        int flows;

        w = getnum(strsep(&cur, ":"), &p, "weight");
        if (w <= 0)
            w = 1;
        w_h = p ? getnum(p+1, &p, "weight_max") : w;
        w_steps = p ? getnum(p+1, &p, "w_steps") : (w_h == w ?1:2);
        len = getnum(strsep(&cur, ":"), &p, "len");
        if (len <= 0)
            len = 1000;
        len_h = p ? getnum(p+1, &p, "len_max") : len;
        l_steps = p ? getnum(p+1, &p, "l_steps") : (len_h == len ? 1 : 2);
        flows = getnum(strsep(&cur, ":"), NULL, "flows");
        if (flows == 0)
            flows = 1;
        DX(5, "weight %d..%d (%d) len %d..%d (%d) flows %d",
            w, w_h, w_steps, len, len_h, l_steps, flows);
        if (w == 0 || w_h < w || len == 0 || len_h < len ||
                flows == 0) {
            DX(5,"wrong parameters %s", fs);
            return;
        }
        n_flows += flows * w_steps * l_steps;
        for (i = 0; i < w_steps; i++) {
            wi = w + ((w_h - w)* i)/(w_steps == 1 ? 1 : (w_steps-1));
            for (j = 0; j < l_steps; j++, n_fs++) {
                struct dn_fs *fs = &c->fs[n_fs].fs; // tentative
                int x;

                li = len + ((len_h - len)* j)/(l_steps == 1 ? 1 : (l_steps-1));
                x = (wi*2048)/li;
                DX(4, "----- fs %4d weight %4d lmax %4d X %4d flows %d",
                    n_fs, wi, li, x, flows);
                if (pass == 0)
                    continue;
                if (c->fs == NULL || c->flowsets <= n_fs) {
                    D("error in number of flowsets");
                    return;
                }
                wsum += wi * flows;
                fs->par[0] = wi;
                fs->par[1] = li > sizeof(struct mbuf) ?
                             li : sizeof(struct mbuf);
                fs->index = n_fs;
                fs->n_flows = flows;
                fs->cur = fs->first_flow = prev==NULL ? 0 : prev->next_flow;
                fs->next_flow = fs->first_flow + fs->n_flows;
                fs->y = x * flows;
                fs->base_y = (prev == NULL) ? 0 : prev->next_y;
                fs->next_y = fs->base_y + fs->y;
                prev = fs;
            }
        }
    }
    c->max_y = prev ? prev->base_y + prev->y : 0;
    c->flows = n_flows;
    c->flowsets = n_fs;
    c->wsum = wsum;
    if (pass == 0)
        return;

    /* now link all flows to their parent flowsets */
    DX(3,"%d flows on %d flowsets max_y %d", c->flows, c->flowsets, c->max_y);
    for (i=0; i < c->flowsets; i++) {
        struct dn_fs *fs = &c->fs[i].fs;
        DX(3, "fs %3d w %5d l %4d flow %5d .. %5d y %6d .. %6d",
            i, fs->par[0], fs->par[1],
            fs->first_flow, fs->next_flow,
            fs->base_y, fs->next_y);
        for (j = fs->first_flow; j < fs->next_flow; j++) {
            struct dn_queue *q = FI2Q(c, j);
            q->fs = &c->fs[i];
        }
    }
}


/* enqueue returns NON-zero when a packet is dropped */
static int
enqueue(struct cfg_s *c, void *_m)
{
    int failure = 0;
    struct mbuf *m = _m;
    struct dn_queue *q = FI2Q(c, m->flow_id);

    if (c->enq) {
        failure = c->enq(c->si, q, m);
        if (! failure && c->single_queue)
            goto inc_length;
        else
            return failure;
    }
    
    if (c->head == NULL)
        c->head = m;
    else
        c->tail->m_nextpkt = m;
    c->tail = m;    
inc_length:
    /*
     * Make sure that each flow is accounted for the number of
     * packets of that flow that are pending, even in case of
     * single_queue schedulers (which do not update the following
     * field)
     */
    q->ni.length++;
    return 0; /* default - success */
}

/* dequeue returns NON-NULL when a packet is available */
static void *
dequeue(struct cfg_s *c)
{
    struct mbuf *m;
    struct dn_queue *q;
    if (c->deq) {
        m = c->deq(c->si);
        if (m != NULL && c->single_queue)
            goto dec_length;
        else
            return m;
    }
    if ((m = c->head)) {
        m = c->head;
        c->head = m->m_nextpkt;
        m->m_nextpkt = NULL;
    }
dec_length:
    q = FI2Q(c, m->flow_id);
    /*
     * Make sure that each flow is accounted (only) for the number
     * of packets of that flow that are pending, even in case of
     * single_queue schedulers (which do not update the following
     * field)
     */
    q->ni.length--;
    return m;
}



/* Create pkt for QosSched starting only from flow_id */
static void *
create_qos_pkt(int flow_id){
    struct mbuf *m_qos;
    struct dn_queue *q_qos;

    q_qos = FI2Q(qossched, flow_id);


    /* Worst Case Fair Index for QoS structure */
    if(q_qos->ni.total_length == 1){ /* 
                                      * the service in now backlogged, 
                                      * start with counters (t1) 
                                      */
        q_qos->ni.allpkt_serv_bklgd = macsal->dequeue;
        q_qos->ni.pkt_serv_bklgd = q_qos->ni.pkts;
    }

    
    /* construct a packet */
    if (freelist) {
        m_qos = tosend = freelist;
        freelist = freelist->m_nextpkt;
    } else {
        m_qos = tosend = calloc(1, sizeof(struct mbuf));
    }
    if (m_qos == NULL)
        return 0;

    m_qos->Q_bklg = q_qos->ni.length + 1; /* update cur. backlog, for T-WFI */
    m_qos->enq_time = macsal->dequeue;
    m_qos->cfg = qossched;
    m_qos->m_nextpkt = NULL;
    m_qos->m_pkthdr.len = sizeof(struct mbuf);
    m_qos->father_flow_id = m_qos->flow_id = flow_id;

    return m_qos;
}

/* MacSal fetch from Qos_Dequeue to MacSAL_enqueue */
static void 
fetch(struct mbuf *m){
    int flow_id;
    struct dn_queue *q_mac;



    flow_id = m->father_flow_id % macsal->flows; /* 
                                                  * simulate flowid remapping 
                                                  * into macsal 
                                                  */
    q_mac = FI2Q(macsal, flow_id);


    /* Worst Case Fair Index */
    if(q_mac->ni.length == 0){ /* 
                                * the service in now backlogged, 
                                * start with counters (t1) 
                                */
        q_mac->ni.allpkt_serv_bklgd = macsal->dequeue;
        q_mac->ni.pkt_serv_bklgd = q_mac->ni.pkts;
    }


#ifdef RELATIVE_FAIRNESS
    /* stat for fairness */

    /* q_mac is becklog */
    if(q_mac->ni.bklgd_start_time == 0 && q_mac->ni.length != 0) 
        q_mac->ni.bklgd_start_time = macsal->_enqueue; 

    /* q_mac is not becklog */
    if(q_mac->ni.bklgd_start_time != 0 && q_mac->ni.length == 0){ 
        q_mac->ni.bklgd_time += (macsal->_enqueue - q_mac->ni.bklgd_start_time);
        q_mac->ni.bklgd_start_time = 0; 
    }
#endif

    

    m->enq_mac_time = macsal->dequeue;
    m->cfg = macsal;
    m->flow_id = flow_id;

    return;
}

/* Create pkt for MacSAL_Sched starting only from flow_id */
static void *
create_macsal_pkt(int qos_flow_id){
    struct mbuf *m_mac;
    int flow_id;
    struct dn_queue *q_qos;
    struct dn_queue *q_mac;

    flow_id = qos_flow_id % macsal->flows;
    q_mac = FI2Q(macsal, flow_id);

    q_qos = FI2Q(qossched, qos_flow_id);

    /* Worst Case Fair Index total */
    if(q_qos->ni.total_length == 1){ /* 
                                      * the service in now backlogged, 
                                      * start with counters (t1) 
                                      */
        q_qos->ni.allpkt_serv_bklgd = macsal->dequeue;
        q_qos->ni.pkt_serv_bklgd = q_qos->ni.pkts;
    }


    /* Worst Case Fair Index for MAC */
    if(q_mac->ni.length == 0){ /* 
                                * the service in now backlogged, 
                                * start with counters (t1) 
                                */
        q_mac->ni.allpkt_serv_bklgd = macsal->dequeue;
        q_mac->ni.pkt_serv_bklgd = q_mac->ni.pkts;
    }

#if RELATIVE_FAIRNESS
    /* stat for fairness */

    /* q_mac is becklog */
    if(q_mac->ni.bklgd_start_time == 0 && q_mac->ni.length != 0) 
        q_mac->ni.bklgd_start_time = macsal->_enqueue;

    /* q_mac is not becklog */
    if(q_mac->ni.bklgd_start_time != 0 && q_mac->ni.length == 0){ 
        q_mac->ni.bklgd_time += (macsal->_enqueue - q_mac->ni.bklgd_start_time);
        q_mac->ni.bklgd_start_time = 0;
    }
#endif
    
    /* construct a packet */
    if (freelist) {
        m_mac = tosend = freelist;
        freelist = freelist->m_nextpkt;
    } else {
        m_mac = tosend = calloc(1, sizeof(struct mbuf));
    }
    if (m_mac == NULL)
        return 0;

    m_mac->Q_bklg = q_qos->ni.total_length + 1; /* cur. backlog, for T-WFI */
    m_mac->enq_time = macsal->dequeue;
    m_mac->enq_mac_time = macsal->dequeue;
    m_mac->cfg = macsal;
    m_mac->m_nextpkt = NULL;
    m_mac->m_pkthdr.len = sizeof(struct mbuf);
    m_mac->flow_id = flow_id;
    m_mac->father_flow_id = qos_flow_id;

    return m_mac;
}

int counter = 0;

static int 
double_enqueue(struct dn_sch_inst *_si, int qos_flow_id)
{
    struct dn_queue *q_qos;
    struct dn_queue *q_mac;
 
    q_qos = FI2Q(qossched,qos_flow_id);
    q_qos->ni.total_length++;
    if(q_qos->ni.total_length  > q_qos->ni.max_length)
        q_qos->ni.max_length = q_qos->ni.total_length ;

    if(macsal->pending < macsal->Q){ /* immediatly mac-sal enqueue */
        macsal->pending++;
        macsal->_enqueue++;

        struct mbuf *m_mac;
        
        m_mac = create_macsal_pkt(qos_flow_id);

        q_mac = FI2Q(macsal,m_mac->flow_id);
 
        if(q_mac->ni.length + 1 > q_mac->ni.max_length)
            q_mac->ni.max_length = q_mac->ni.length + 1 ;

        ND("MAC flow_id %d", m_mac->flow_id);

        ND("mac_enqueue, pending %d flow %5d fset %d flowset len %4d",
            macsal->pending, m_mac->flow_id, macsal->cur_fs,
            m_mac->m_pkthdr.len);

        return enqueue(macsal,m_mac);

    } else { /* enqueue in QoS scheduler */
        qossched->pending++;
        qossched->_enqueue++;

        struct mbuf *m_qos = create_qos_pkt(qos_flow_id);

        ND("QoS enq, pending %d flow %5d fset %d len %4d",
            qossched->pending, m_qos->flow_id, qossched->cur_fs,
            m_qos->m_pkthdr.len);

        return enqueue(qossched,m_qos);
    }
}

static int
double_dequeue(struct dn_sch_inst *_si)
{
    int qos_flow_id;
    struct mbuf *m_mac;
    struct dn_queue *q;
    struct dn_queue *q_mac;
    int queueing_delay,queueing_delay_mac;
    double wfi;
    double twfi;

    /* 
     * extract the macsal packet dequeue and recover 
     * the qos one associated to return 
     */
    if ((m_mac = dequeue(macsal))) {
        macsal->dequeue++;
        qos_flow_id = m_mac->father_flow_id;

        /* STATS queueing delay */
        queueing_delay = macsal->dequeue - m_mac->enq_time;
        queueing_delay_mac = macsal->dequeue - m_mac->enq_mac_time;

        q = FI2Q(qossched, qos_flow_id);
        q_mac = FI2Q(macsal, m_mac->flow_id);

        q->ni.total_length--;

        ND("MAC fl_id %d, weight %d", m_mac->flow_id, q_mac->fs->fs.par[0]);
        ND("QoS fl_id %d, weight %d totla_l %d", 
            m_mac->father_flow_id, q->fs->fs.par[0], q->ni.total_length);


        /* update global stats */
        /* max time */
        if (queueing_delay > q->ni.q_time_max)
            q->ni.q_time_max = queueing_delay;
        /* min time */
        if (queueing_delay < q->ni.q_time_min || q->ni.q_time_min == 0)
            q->ni.q_time_min = queueing_delay;
        /* avg time */
        q->ni.q_avg_time += queueing_delay;
        /* time variance (at the end divided for all the packets dequeued)*/
        q->ni.q_time_square += (queueing_delay*queueing_delay);

        /* Update MAC-SAL queueing-delay stats */
        /* max time */
        if (queueing_delay_mac > q_mac->ni.q_time_max)
            q_mac->ni.q_time_max = queueing_delay_mac;
        /* min time */
        if (queueing_delay_mac < q_mac->ni.q_time_min ||
           q_mac->ni.q_time_min == 0)
            q_mac->ni.q_time_min = queueing_delay_mac;
        /* avg time */
        q_mac->ni.q_avg_time += queueing_delay_mac;
        /* time variance (at the end divided for all the packets dequeued)*/
        q_mac->ni.q_time_square += (queueing_delay_mac*queueing_delay_mac);


        q->ni.pkts++;
        q_mac->ni.pkts++;
        q_mac->ni.wind_pkts++;

        /* update stats for T-WFI*/
        twfi = queueing_delay -
               m_mac->Q_bklg / ( (double)q->fs->fs.par[0] / qossched->wsum);

        if(twfi > t_wfi && q_mac->fs->fs.par[0] == macsal->max_flow_w && 
           q->fs->fs.par[0] == qossched->max_flow_w )
               t_wfi = twfi;
        /* 
         * TIME Worst Case Fair Index intermediate, 
         * good QoS flow mapped in bad MAC flow
         */
        if(twfi > t_wfi_interm && q_mac->ni.ploss == 5 &&
           q->fs->fs.par[0] == qossched->max_flow_w )
            t_wfi_interm = twfi;


        /* update stats for global B-WFI */
        if(q->ni.total_length == 0){ 
            /* 
             * the service in no more backlogged starting 
             * from now, stops with counters (t2) 
             */
            q->ni.allpkt_serv_bklgd = macsal->dequeue - q->ni.allpkt_serv_bklgd;
            q->ni.pkt_serv_bklgd = q->ni.pkts - q->ni.pkt_serv_bklgd;
            
            wfi =((double)q->fs->fs.par[0]/qossched->wsum)*
                 q->ni.allpkt_serv_bklgd - q->ni.pkt_serv_bklgd;

            if(wfi > b_wfi_qos && q_mac->fs->fs.par[0] == macsal->max_flow_w && 
               q->fs->fs.par[0] == qossched->max_flow_w)
                b_wfi_qos = wfi;
            
            /* 
             * Worst Case Fair Index QoS intermediate, 
             * good QoS flow mapped in bad MAC flow
             */
            if(wfi > b_wfi_qos_interm && q_mac->ni.ploss == 5 && 
               q->fs->fs.par[0] == qossched->max_flow_w){
                b_wfi_qos_interm = wfi;
            }
        }


        /* update stats for MAC-SAL B-WFI */
        if(q_mac->ni.length == 0){ 
            /* 
             * the service in no more backlogged starting 
             * from now, stops with counters (t2) 
             */
            q_mac->ni.allpkt_serv_bklgd = macsal->dequeue - 
                                          q_mac->ni.allpkt_serv_bklgd;

            q_mac->ni.pkt_serv_bklgd = q_mac->ni.pkts - 
                                       q_mac->ni.pkt_serv_bklgd;

            wfi =((double)q_mac->fs->fs.par[0]/macsal->wsum)*
                 q_mac->ni.allpkt_serv_bklgd - q_mac->ni.pkt_serv_bklgd;
            if(wfi > b_wfi)
                b_wfi = wfi;
        }

        double_drop(macsal,m_mac); 
        macsal->drop--; 
    } else { /* packet allocation failure */
        return -1;
    }

    /* refill MAC-SAL if needed */
    if(macsal->pending == macsal->Q && qossched->pending > 0){ 
        ND("macsal full, deq from macsal: pending = %d, Q = %d, qos_pend = %d", 
            macsal->pending, macsal->Q, qossched->pending);
            m_mac = dequeue(qossched);
            qossched->dequeue++;
            qossched->pending--;
            macsal->_enqueue++;
        
            fetch(m_mac);
            qossched->drop--;   

            ND("enqueue in mac-sal, pending %d flow %5d len %4d",
                macsal->pending, m_mac->flow_id, macsal->cur_fs);
            enqueue(macsal,m_mac);
    } else { 
        /* 
         * Extract a packet from MAC-SAL sched without refilling it 
         */
        ND("macsal not full pend %d QoS empty pend %d no quick mac_enqueue", 
            macsal->pending, qossched->pending);
        macsal->pending--;
    }
    return qos_flow_id;
}

/* to set the ploss per flow */
static void
parse_ploss_per_flow(const char *fs)
{
    char *s, *cur, *next;
    int i = 0;
    struct dn_queue *q_mac;

    s = strdup(fs);

    for (next = s; (cur = strsep(&next, ":"));) {
        if(i < macsal->flows){
            q_mac = FI2Q(macsal, i++);
            q_mac->ni.ploss = atoi(cur);
        }
        else
            D("too much ploss then mac flows");
    }
    if(i < macsal->flows){
        D("too few ploss then mac flows");
        for(;i < macsal->flows;i++){
            q_mac = FI2Q(macsal, i);
            q_mac->ni.ploss = 9; /* Default good class */
        }
    }
}

/* to set the ploss per flow AUTOMATICALY */
static void
compute_initial_ploss()
{
    int i = 0;
    struct dn_queue *q_mac;

    /* 
     * in this version we have a static loss. recompute ploss is high 
     * and the entire program stops before change it 
     */
    recompute_ploss_interval = macsal->loops;

    for(i=0;i<macsal->flows;i++){
        q_mac = FI2Q(macsal, i);

        /* static p_loss distribution among flows */
        q_mac->ni.ploss = i % 10;


        if(macsal->weight_distrib == 0){ /* CONSTANT Weight */
            q_mac->fs = &macsal->fs[0]; /* all flows in the same flowset */
            q_mac->fs->fs.par[0] = 1;
        }                   
        if(macsal->weight_distrib == 1){ /* LINEAR weight */
            /* all flows of the same ploss in the same flowset */
            q_mac->fs = &macsal->fs[q_mac->ni.ploss]; 
            q_mac->fs->fs.par[0] = q_mac->ni.ploss + 1; 
        }
        if(macsal->weight_distrib == 2){ /* EXPONENTIAL weight */
            /* all flows of the same ploss in the same flowset */
            q_mac->fs = &macsal->fs[q_mac->ni.ploss]; 
            q_mac->fs->fs.par[0] = pow(2,q_mac->ni.ploss);
        }
        if(macsal->weight_distrib == 3){ /*ANALOGICAL weight*/
            /* all flows of the same ploss in the same flowset */
            q_mac->fs = &macsal->fs[q_mac->ni.ploss]; 
            q_mac->fs->fs.par[0] = (p_loss[q_mac->ni.ploss]) * 999 + 1; 
            if(q_mac->fs->fs.par[0] == 999)
                q_mac->fs->fs.par[0] = 1000;
            if(q_mac->fs->fs.par[0] < 800 && q_mac->fs->fs.par[0] > 1)
                q_mac->fs->fs.par[0] = (int)(q_mac->fs->fs.par[0] / 10);
            if(q_mac->fs->fs.par[0] <= 900 && q_mac->fs->fs.par[0] >= 800)
                q_mac->fs->fs.par[0] =  400;
        }
    }
}

static int
double_new_sched(struct dn_sch_inst *_si)
{
    int i;
    int ac = _si->sched->fs->fs.n_args;
    char *const *av = _si->sched->fs->fs.args;

    char *ploss = NULL;

    moduledata_t *qos = NULL;
    moduledata_t *mac = NULL;

    /* define the structure and the parameters for the MAC-SAL */
    macsal = calloc(1, sizeof(struct cfg_s));
    qossched = calloc(1, sizeof(struct cfg_s));
    macsal->si_len = sizeof(struct dn_sch_inst);
    qossched->si_len = sizeof(struct dn_sch_inst);
    macsal->q_len = sizeof(struct dn_queue);
    qossched->q_len = sizeof(struct dn_queue);

    extern moduledata_t *_g_dn_fifo;        
    extern moduledata_t *_g_dn_wf2qp;
    extern moduledata_t *_g_dn_rr;
    extern moduledata_t *_g_dn_qfq;
    extern moduledata_t *_g_dn_qfqp;
    extern moduledata_t *_g_dn_kps;

    macsal->lmin = macsal->lmax = qossched->lmin = qossched->lmax 
                 = sizeof(struct mbuf);
    macsal->flows = qossched->flows = 1;
    macsal->max_flow_w = qossched->max_flow_w = 0;
    qossched->flowsets = 1;
    macsal->flowsets = 11; /* the actual number of pLoss Class */
    macsal->Q = 1;
    macsal->name = qossched->name = "null";
    macsal->weight_distrib = 0;
    macsal->intgr_th = 0;

    ac--; av++;
    while (ac > 1) {
        if (!strcmp(*av, "-qos")) {
            if (!strcmp(av[1], "rr"))
                qos = _g_dn_rr;
            else if (!strcmp(av[1], "wf2qp"))
                qos = _g_dn_wf2qp;
            else if (!strcmp(av[1], "fifo"))
                qos = _g_dn_fifo;
            else if (!strcmp(av[1], "qfq"))
                qos = _g_dn_qfq;
            else if (!strcmp(av[1], "qfq+"))
                qos = _g_dn_qfqp;
            else if (!strcmp(av[1], "kps"))
                qos = _g_dn_kps;
            else
                qos = NULL;
            DX(3, "using qos scheduler %s", qos->name);
        } else if (!strcmp(*av, "-mac")) {
            if (!strcmp(av[1], "rr"))
                mac = _g_dn_rr;
            else if (!strcmp(av[1], "wf2qp"))
                mac = _g_dn_wf2qp;
            else if (!strcmp(av[1], "fifo"))
                mac = _g_dn_fifo;
            else if (!strcmp(av[1], "qfq"))
                mac = _g_dn_qfq;
            else if (!strcmp(av[1], "qfq+"))
                mac = _g_dn_qfqp;
            else if (!strcmp(av[1], "kps"))
                mac = _g_dn_kps;
            else
                mac = NULL;
            DX(3, "using mac scheduler %s", mac->name);
        } else if (!strcmp(*av, "-n")) {
            macsal->loops = getnum(av[1], NULL, av[0]);
        } else if (!strcmp(*av, "-lenmac")) {
            macsal->lmin = getnum(av[1], NULL, av[0]);
            macsal->lmax = macsal->lmin;
            DX(3, "setting lenmac to %d", macsal->lmax);
        } else if (!strcmp(*av, "-qmac")) {
            macsal->Q = getnum(av[1], NULL, av[0]);
            DX(3, "setting qmac to %d", macsal->Q);
        } else if (!strcmp(*av, "-stats")) {
            macsal->print_stats_level = getnum(av[1], NULL, av[0]);
            DX(3, "setting print stats to %d", macsal->print_stats_level);
        } else if (!strcmp(*av, "-intgr_th")) {
            macsal->intgr_th = getnum(av[1], NULL, av[0]);
            DX(3, "set intgr_th to emulate G/B state");
        } else if (!strcmp(*av, "-flows")) {
            qossched->flows = getnum(av[1], NULL, av[0]);
            DX(3, "setting macflows to %d", qossched->flows);
        } else if (!strcmp(*av, "-flowsets")) {
            parse_flowsets(qossched, av[1], 0);
            DX(3, "setting flowsetsmac to %d", qossched->flowsets);
        } else if (!strcmp(*av, "-flowsmac")) {
            macsal->flows = getnum(av[1], NULL, av[0]);
            DX(3, "setting macflows to %d", macsal->flows);
        } else if (!strcmp(*av, "-flowsetsmac")) {
            parse_flowsets(macsal, av[1], 0);
            DX(3, "setting flowsetsmac to %d", macsal->flowsets);
        } else if (!strcmp(*av, "-ploss")) {
            ploss = av[1];
        } else if (!strcmp(*av, "-wdistr")) {
            macsal->weight_distrib = getnum(av[1], NULL, av[0]);
        } else {
            DX(0,"option %s not recognised for macsal, ignore", *av);
        }
        ac -= 2; av += 2;
    } /* set the eventually wrong MAC-SAL parameter */
    if (macsal->flows <= 0)
        macsal->flows = 1;
    if (macsal->flowsets <= 0)
        macsal->flowsets = 1;
    if (macsal->Q <= 0)
        macsal->Q = 10;
    if (qossched->flows <= 0)
        qossched->flows = 1;
    if (qossched->flowsets <= 0)
        qossched->flowsets = 1;
    if (macsal->intgr_th < 0)
        macsal->intgr_th = 0;

    if (macsal->weight_distrib < 0 || macsal->weight_distrib > 3)
        macsal->weight_distrib = 0;

    macsal->N = _si->sched->fs->fs.qos_flows;
    macsal->M = macsal->flows;

    if(macsal->M > macsal->N){
        D("macsal flows %d are more then qos flows %d", macsal->M , macsal->N);
        exit(1);
    }

    /* Define the QoS algorithm */
    if(qos)
        p_qos = qos->p;
    else{
        D("no -qos algo found");
        exit(1);
    }

    /* Define the MAC-SAL algorithm */
    if(mac)
        p_mac = mac->p;
    else{
        D("no -mac algo found");
        exit(1);
    }
           
    if (qos == NULL || qos == _g_dn_fifo)
        qossched->single_queue = 1;

    if (qos) {
        qossched->enq = p_qos->enqueue;
        qossched->deq = p_qos->dequeue;
        qossched->si_len += p_qos->si_datalen;
        qossched->q_len += p_qos->q_datalen;
        qossched->schk_len += p_qos->schk_datalen;
    }

    /* allocate queues, flowsets and one scheduler for QoS-sched */
    qossched->q = calloc(qossched->flows, qossched->q_len);
    qossched->fs = calloc(qossched->flowsets, sizeof(struct dn_fsk));
    qossched->si = calloc(1, qossched->si_len);
    qossched->sched = calloc(qossched->flows, qossched->schk_len);
    if (qossched->q == NULL || qossched->fs == NULL) {
        D("error allocating memory for flows");
        exit(1);
    }
    qossched->si->sched = qossched->sched;
    if (p_qos) {
        if (p_qos->config)
            p_qos->config(qossched->sched);
        if (p_qos->new_sched)
            p_qos->new_sched(qossched->si);
    }

    /* parse_flowsets links queues to their flowsets */
    parse_flowsets(qossched, av[1], 1);
    /* complete the work calling new_fsk for qossched */
    for (i = 0; i < qossched->flowsets; i++) {
        if (qossched->fs[i].fs.par[1] == 0)
            qossched->fs[i].fs.par[1] = sizeof(struct mbuf); /*default pkt len*/
        qossched->fs[i].sched = qossched->sched;
        if (p_qos && p_qos->new_fsk)
            p_qos->new_fsk(&qossched->fs[i]);
    }

    for (i = 0; i < qossched->flows; i++) {
        struct dn_queue *q_qos = FI2Q(qossched, i);

        if (q_qos->fs == NULL)
            q_qos->fs = &qossched->fs[0]; /* XXX */
        q_qos->_si = qossched->si;
        if (p_qos && p_qos->new_queue)
            p_qos->new_queue(q_qos);
    }

    for(i = 0; i < qossched->flows; i++){
        struct dn_queue *q_qos = FI2Q(qossched, i);

        if(q_qos->fs->fs.par[0] > qossched->max_flow_w)
            qossched->max_flow_w = q_qos->fs->fs.par[0];
    }

    DX(1,"Qos sched new_sched() done");

    if (mac == NULL || mac == _g_dn_fifo)
        macsal->single_queue = 1;

    if (mac) {
        macsal->enq = p_mac->enqueue;
        macsal->deq = p_mac->dequeue;
        macsal->si_len += p_mac->si_datalen;
        macsal->q_len += p_mac->q_datalen;
        macsal->schk_len += p_mac->schk_datalen;
    }

    /* allocate queues, flowsets and one scheduler for mac-sal */
    macsal->q = calloc(macsal->flows, macsal->q_len);
    macsal->fs = calloc(macsal->flowsets, sizeof(struct dn_fsk));
    macsal->si = calloc(1, macsal->si_len);
    macsal->sched = calloc(macsal->flows, macsal->schk_len);
    if (macsal->q == NULL || macsal->fs == NULL) {
        D("error allocating memory for flows");
        exit(1);
    }
    macsal->si->sched = macsal->sched;
    if (p_mac) {
        if (p_mac->config)
            p_mac->config(macsal->sched);
        if (p_mac->new_sched)
            p_mac->new_sched(macsal->si);
    }

    /* parse_flowsets links queues to their flowsets */
    parse_flowsets(macsal, av[1], 1);
    /* complete the work calling new_fsk for macsal */
    for (i = 0; i < macsal->flowsets; i++) {
        if (macsal->fs[i].fs.par[1] == 0)
            macsal->fs[i].fs.par[1] = sizeof(struct mbuf); /*default pkt len*/
        macsal->fs[i].sched = macsal->sched;
        if (p_mac && p_mac->new_fsk)
            p_mac->new_fsk(&macsal->fs[i]);
    }

    if(ploss != NULL){
        parse_ploss_per_flow(ploss);
    }
    else compute_initial_ploss();

    for (i = 0; i < macsal->flows; i++) {
        struct dn_queue *q_mac = FI2Q(macsal, i);
        macsal->wsum +=  q_mac->fs->fs.par[0];
    }

    for (i = 0; i < macsal->flows; i++) {
        struct dn_queue *q_mac = FI2Q(macsal, i);

        if (q_mac->fs == NULL)
            q_mac->fs = &macsal->fs[0]; /* XXX */
        q_mac->_si = macsal->si;
        if (p_mac && p_mac->new_queue)
            p_mac->new_queue(q_mac);
    }

    for(i = 0; i < macsal->flows; i++){
        struct dn_queue *q_mac = FI2Q(macsal, i);

        if(q_mac->fs->fs.par[0] > macsal->max_flow_w)
            macsal->max_flow_w = q_mac->fs->fs.par[0];
    }

    DX(1,"Mac sched new_sched() done");

    return 0;
}

/* update and print statistics at the end of the run */
static void
double_finalprint_stats(struct dn_sch_inst *_si)
{

    struct dn_queue *q;
    struct dn_queue *q_qos;
    double avg;
    double sigma;
    int i;
    int Ap = 0;
    int Ao = macsal->dequeue;
    int Ae;
    double thr;
    int max_delay = 0;
    int max_delay_interm_flow = 0; /* available only if weight are static! */
    int max_delay_best_flow = 0; /* available only if weight are static! */

    double min_pkt_norm = macsal->loops; /*start from max to achive min*/
    double max_pkt_norm = 0;
    double avg_pkt_norm = 0;
    double wfi = 0;


    int min_blog_time = macsal->loops; /*start from max to achive min*/

    if(macsal->print_stats_level == 0)
	    return;

    if(macsal->print_stats_level > 1){

        /* Queueing Delay total (QoS + MAC-SAL) */
        for(i = 0; i < qossched->flows; i++){
            q_qos = FI2Q(qossched, i);
            q    = FI2Q(macsal, i % macsal->flows);
            if(q->fs->fs.par[0] == macsal->max_flow_w &&  
               q_qos->fs->fs.par[0] == qossched->max_flow_w ) {
                if(((double)q_qos->ni.q_time_max * 8.0 * 
                    sizeof(struct mbuf))/bw > q_del)
                    q_del = ((double)q_qos->ni.q_time_max * 8.0 * 
                             sizeof(struct mbuf))/bw;
            }
        }

        /* 
         * stats for Worst Case Fair Index when (t2) 
         * falls on the end of experiment 
         */
        for(i = 0; i < qossched->flows; i++){
            q_qos = FI2Q(qossched, i);
            /* the service in no more backlogged starting from now */
            if(q_qos->ni.total_length != 0){
                q_qos->ni.allpkt_serv_bklgd = macsal->dequeue - 
                                              q_qos->ni.allpkt_serv_bklgd;

                q_qos->ni.pkt_serv_bklgd = q_qos->ni.pkts - 
                                           q_qos->ni.pkt_serv_bklgd;
            }
        }

        /* calculate B-WFI for each QoS flow */
        for(i = 0; i < qossched->flows; i++){
            q_qos = FI2Q(qossched, i);
            q    = FI2Q(macsal, i % macsal->flows);
            wfi =((double)q_qos->fs->fs.par[0]/qossched->wsum)*
                 q_qos->ni.allpkt_serv_bklgd - q_qos->ni.pkt_serv_bklgd;

            if(wfi > b_wfi_qos && q->fs->fs.par[0] == macsal->max_flow_w && 
                q_qos->fs->fs.par[0] == qossched->max_flow_w)
                b_wfi_qos = wfi;
        }
                    
        /* 
         * stats for Worst Case Fair Index when (t2) 
         * falls on the end of experiment 
         */
        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
            /* the service in no more backlogged starting from now */
            if(q->ni.length != 0){ 
                q->ni.allpkt_serv_bklgd = macsal->dequeue -
                                          q->ni.allpkt_serv_bklgd;
                q->ni.pkt_serv_bklgd = q->ni.pkts - q->ni.pkt_serv_bklgd;
            }
        }

        /* calculate B-WFI for each mac flow */
        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
            wfi = ((double)q->fs->fs.par[0]/macsal->wsum)*
                  q->ni.allpkt_serv_bklgd - q->ni.pkt_serv_bklgd;
            if(wfi > b_wfi)
                b_wfi = wfi;
        }

        /* calculate bklgd_time for flow always backlogged at run_time */
        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
            if(q->ni.bklgd_time == 0)
                q->ni.bklgd_time = macsal->_enqueue - q->ni.bklgd_start_time;
        }

        /* calculate min_blog_time for fairness stats */
        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
            if(q->ni.bklgd_time < min_blog_time)
                min_blog_time = q->ni.bklgd_time;
        }

        /* calculate pkts_all_blog for each flow */
        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
                        
            q->ni.pkts_all_blog = q->ni.wind_pkts * 
                                  min_blog_time/q->ni.bklgd_time;
        }

        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
            if((q->ni.pkts_all_blog != 0) && 
              ((double)q->ni.pkts_all_blog / q->fs->fs.par[0]) < min_pkt_norm)
                min_pkt_norm = (double)q->ni.pkts_all_blog / q->fs->fs.par[0] ;
            if((q->ni.pkts_all_blog != 0) && 
              ((double)q->ni.pkts_all_blog / q->fs->fs.par[0]) > max_pkt_norm)
                max_pkt_norm = (double)q->ni.pkts_all_blog / q->fs->fs.par[0] ;
        }

        avg_pkt_norm = max_pkt_norm - min_pkt_norm;

        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
            if(q->ni.q_time_max > max_delay)
                max_delay = q->ni.q_time_max;
            if(q->ni.q_time_max > max_delay_interm_flow && q->ni.ploss == 5)
                max_delay_interm_flow = q->ni.q_time_max;
            if(q->ni.q_time_max > max_delay_best_flow && 
               q->fs->fs.par[0] == macsal->max_flow_w)
                max_delay_best_flow = q->ni.q_time_max;
        }

        /* accumulate stats for the last period */
        for(i=0;i<macsal->flows;i++){
            q = FI2Q(macsal, i);
            q->ni.loss_pkts += q->ni.wind_pkts*(1.0 - p_loss[q->ni.ploss]);
        }

        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
            if(macsal->intgr_th != 0 && q->fs->fs.par[0] < macsal->intgr_th){
            /*
             * if weight flow is under threshold we consider it inactive
             * to simulate integrated scheduler Bad state
             */
            Ao -= q->ni.wind_pkts;
        }
        else
            Ap += q->ni.loss_pkts;
        }
        Ae = Ao - Ap;
        thr = (double)Ae/Ao;

        avg=0;
        for(i = 0; i < macsal->flows; i++){
            q = FI2Q(macsal, i);
            avg += (double)q->ni.q_avg_time / q->ni.pkts;
        }
        avg = (double)avg / macsal->flows;


        q_del_mac_best_flow = (double)max_delay_best_flow * 8.0 * 
                          sizeof(struct mbuf)/bw;
        q_del_mac_interm_flow = (double)max_delay_interm_flow * 8.0 * 
                         sizeof(struct mbuf)/bw;

        t_wfi *= 8.0 * sizeof(struct mbuf)/bw;
        t_wfi_interm *= 8.0 * sizeof(struct mbuf)/bw;

        if(macsal->print_stats_level > 2){
            if(macsal->print_stats_level > 2){
                printf("%5.3f  %5.3f  %5.3f  %5.3f  %5.3f  ",
                       q_del, b_wfi_qos, b_wfi_qos_interm, t_wfi, t_wfi_interm);
                return ;
            }
            D("Q  thr  delay10-1  delay10-7  avg_delay  fairness_index"
               "WorsFairIndx  QueueingDelay(all)  QueueingDelay_seven(mac)" 
               "QueueingDelay_two(mac)  exec_time");
        }
        printf("%d  %.3f  %d  %d  %5.3f  %5.3f  %5.3f  %5.3f  %5.3f  %5.3f  ",
                macsal->Q, thr, max_delay_interm_flow, max_delay_best_flow, 
                avg, avg_pkt_norm, b_wfi, q_del, q_del_mac_best_flow, 
                q_del_mac_interm_flow);

        return ;
    }


    D("QOS STATS:");
    for(i = 0; i < qossched->flows; i++){
        q = FI2Q(qossched, i);
        if(q->ni.pkts > 0){
            avg = (double)q->ni.q_avg_time / q->ni.pkts;
            sigma = abs(q->ni.q_time_square/ (double)q->ni.pkts - avg*avg);
        }
        D("flow %d, pkts %d, min %d, max %d, avg %.2f, vrc %.2f, max_len %d",
          i, q->ni.pkts, q->ni.q_time_min, q->ni.q_time_max, avg, sigma, 
          q->ni.max_length);
    }

    D("MAC STATS (about QoS performances):");
    for(i = 0; i < macsal->flows; i++){
        q = FI2Q(macsal, i);
        if(q->ni.pkts > 0){
            avg = (double)q->ni.q_avg_time / q->ni.pkts;
            sigma = abs(q->ni.q_time_square/ (double)(q->ni.pkts - avg*avg));
        }
        D("flow %d, pkts %d, min %d, max %d, avg %.2f, vrc %.2f, max_len %d",
          i, q->ni.pkts, q->ni.q_time_min, q->ni.q_time_max, avg, sigma, 
          q->ni.max_length);
    }

    D("MAC STATS:");
    /*accumulate stats for the last period*/
    for(i=0;i<macsal->flows;i++){
        q = FI2Q(macsal, i);
        q->ni.loss_pkts += q->ni.wind_pkts * (1.0 - p_loss[q->ni.ploss]);
    }

    for(i = 0; i < macsal->flows; i++){
        q = FI2Q(macsal, i);
        if(macsal->intgr_th != 0 && q->fs->fs.par[0] < macsal->intgr_th){
            /*
             * if weight flow is under threshold we consider it inactive
             * to simulate integrated scheduler Bad state
             */
            Ao -= q->ni.wind_pkts;
        }
        else
            Ap += q->ni.loss_pkts;
        D("flow %d, pkts %d, loss_class %d, pack_loss %d",
               i, q->ni.pkts, q->ni.ploss, q->ni.loss_pkts);
    }
    Ae = Ao - Ap;
    D("Traffic offered: %d, trasmitted: %d",Ao, Ae);
    thr = (double)Ae/Ao;
    D("Throughput = %.3f", thr);

}


/*
 * DOUBLE scheduler descriptor
 * contains the type of the scheduler, the name, the size of the
 * structures and function pointers.
 */
static struct dn_alg double_desc = {
    .type = DN_DOUBLE,
    .name = "DOUBLE",
    .flags = DN_MULTIQUEUE,

    .si_datalen = sizeof(struct dn_queue),
    .q_datalen = sizeof(struct dn_queue),

    .enqueue_wrapper = double_enqueue,
    .dequeue_wrapper = double_dequeue,

    .new_sched = double_new_sched,
    .finalprint_stats = double_finalprint_stats,
};


DECLARE_DNSCHED_MODULE(dn_double, &double_desc);
